
import anndata as ad
import scvelo as scv
import scanpy as sc
import pandas as pd
import numpy as np
import os

def cluster_velocities(h5ad_path,
                       resolution=0.05,
                       prefix="vcluster",
                       cluster_key = "velocity_clusters"):
    adata = ad.read_h5ad(h5ad_path) if type(h5ad_path) == str else h5ad_path

    v_dat = adata.layers["velocity"]
    v = sc.AnnData(v_dat)
    sc.pp.pca(v)
    sc.pp.neighbors(v)
    sc.tl.umap(v)
    sc.tl.leiden(v, resolution=resolution)


    v.obs["leiden"] = prefix + v.obs["leiden"].astype(str)
    v_clusts = v.obs["leiden"].to_frame()
    v_clusts = v_clusts.set_index(adata.obs_names.values)
    adata.obs[cluster_key] = v_clusts
    print("Velocity clusters found: ")
    value_counts = adata.obs[cluster_key].value_counts()
    print(value_counts)
    significant_vclusters = value_counts[value_counts > (0.01 * adata.n_obs)]
    print("Found " + str(len(significant_vclusters)) + " significant velocity clusters (clusters with more than 1% of total cells.)")


    return(adata)


def recover_dynamics_clusters(adata, 
                              return_adata = False,
                              cluster_key = "velocity_clusters",
                              **kwargs):
    if type(adata) == str:
        adata = ad.read_h5ad(adata)

    vclust_likelihoods = []
    vclusts = adata.obs[cluster_key].unique()
    for vclust in vclusts:
        vclust_cell_idx = np.where(adata.obs[cluster_key] == vclust)
        ad_clust = adata[vclust_cell_idx].copy()
        scv.tl.recover_dynamics(ad_clust, **kwargs)

        top_likelihoods_clust = ad_clust.var['fit_likelihood']
        top_likelihoods_clust.rename(vclust, inplace=True)
        vclust_likelihoods.append(top_likelihoods_clust)


    likelihoods_df = pd.concat(vclust_likelihoods, axis=1)
    cluster_df = pd.DataFrame(adata.obs[cluster_key])

    if (return_adata == True):
        adata.uns["velocity_cluster_likelihoods"] = likelihoods_df
        return(adata)
        
    return([likelihoods_df, cluster_df])

def plot_sender_influence_velocity_python(adata, 
                                  sender_adata = None,
                                  velocity_clusters = None, 
                                  velocity_cluster_key = "velocity_clusters", 
                                  top_n_ligands=3, 
                                  vkey = "velocity",
                                  basis="umap",
                                  arrow_color="black",
                                  dpi=300,
                                  figsize = None,
                                  size=15,
                                  density=None,
                                  arrow_length=2,
                                  arrow_size=3,
                                  sender_palette = "OrRd",
                                  filename = "entrain_velocity_influence.png",
                                  vector_type = "stream",
                                  signif_cluster_pct_cutoff = 0.01,
                                  frameon=None,
                                  **kwargs):
    import scvelo as scv
    import matplotlib.pyplot as plt
    import math
    if type(adata) == str: adata = ad.read_h5ad(adata)
    if type(sender_adata) == str: sender_adata = ad.read_h5ad(sender_adata) 
    
    if velocity_clusters:
        if np.all(np.isin(velocity_clusters, adata.obs[velocity_cluster_key].unique())):
            clusters = velocity_clusters
        else:
            raise ValueError(velocity_clusters + "not found in values of column: " + velocity_cluster_key)
    elif signif_cluster_pct_cutoff > 0.0:
        vcluster_counts = adata.obs[velocity_cluster_key].value_counts()
        clusters = vcluster_counts[vcluster_counts > (signif_cluster_pct_cutoff * adata.n_obs)].index.to_numpy()
    else:    
         clusters = adata.obs[velocity_cluster_key].unique()          
    
    velocity_basis=vkey + "_" + basis
    if velocity_basis not in adata.obsm.keys():
        scv.tl.velocity_embedding(adata, basis = basis, vkey = vkey)

    if vector_type not in ["stream", "grid"]:
        raise ValueError("vector_type must be one of \"stream\" or \"grid\".")

    adjust_for_stream = True if vector_type == "stream" else False

    fig = plt.figure()
    if sender_adata:
        
        nrows=2
        ncols=len(clusters)
        ax = []
        
        figsize = figsize if figsize is not None else (8*ncols, 16)
        gs = plt.GridSpec(nrows,
                          ncols,
                          plt.figure(figsize=figsize)
                          ) 
        for i, velocity_cluster in enumerate(clusters):
            velocity_cluster = str(velocity_cluster)
            print("Plotting velocities and ligands for cluster: " + velocity_cluster)
            cells = adata.obs.loc[adata.obs[velocity_cluster_key] == velocity_cluster].index
            cells_idx = np.where(adata.obs_names.isin(cells))[0].tolist()
        
            X_grid, V_grid = get_embedding_subset_grid(adata = adata,
                                                       cells_idx = cells_idx,
                                                       basis = basis,
                                                       vkey = vkey,
                                                       adjust_for_stream=adjust_for_stream
                                                       )
            
            if vector_type == "grid":
                ax.append(
                    scv.pl.velocity_embedding_grid(adata, 
                                                      palette=sender_palette,
                                                      color = velocity_cluster_key,
                                                      dpi=300, 
                                                      size=size,
                                                      arrow_length=arrow_length,
                                                      arrow_size=arrow_size,
                                                      density=density,
                                                      arrow_color=arrow_color, 
                                                      show=False,
                                                      ax=plt.subplot(gs[0,i]),
                                                      X_grid = X_grid,
                                                      V_grid = V_grid,
                                                      autoscale=True,
                                                      frameon=frameon,
                                                      title = velocity_cluster
                                                      )   
                    )
            elif vector_type == "stream":
                ax.append(
                    scv.pl.velocity_embedding_stream(adata, 
                                                      palette=sender_palette,
                                                      color = velocity_cluster_key,
                                                      dpi=300, 
                                                      size=size,
                                                      arrow_size=arrow_size,
                                                      density=density,
                                                      arrow_color=arrow_color, 
                                                      show=False,
                                                      ax=plt.subplot(gs[0,i]),
                                                      X_grid = X_grid,
                                                      V_grid = V_grid,
                                                      frameon=frameon,
                                                      title = velocity_cluster
                                                      )   
                    )       
            entrain_dat = adata.uns["entrain_velocity_ligands"]
            ligand_importances = entrain_dat["vclust_ligand_importances"][velocity_cluster].fillna(0)
            ligand_importances.index = entrain_dat["vclust_ligand_importances"]["ligand"]
            ligand_importances = ligand_importances.sort_values(ascending=False)
            ligands = ligand_importances.index.values
            
            sender_adata.obs["influence_"+velocity_cluster] = get_sender_influence(sender_adata, ligands)

            ax.append(
                sc.pl.umap(sender_adata,
                       color="sender_influence",
                       label="on data",
                       palette=sender_palette,
                       title=velocity_cluster,
                       show=False,
                       ax=plt.subplot(gs[1,i])
                       )
                )
            
        ax[0].figure.savefig(filename, dpi=dpi)

    else:
        nrows = 1 if len(clusters) < 5 else 1 + (math.floor(len(clusters) / 4) )
        ncols = len(clusters) if len(clusters) < 4 else 4
        ax = []
        
        figsize = figsize if figsize is not None else (8*ncols, 8*nrows)
        gs = plt.GridSpec(nrows,
                          ncols,
                          plt.figure(figsize=figsize)
                          ) 
        for i, velocity_cluster in enumerate(clusters):
            velocity_cluster = str(velocity_cluster)
            gs_row = math.floor(i/4)
            gs_col = i % 4
            
            print("Plotting velocities and ligands for cluster: " + velocity_cluster)
            if velocity_cluster not in adata.obs[velocity_cluster_key].unique():
                raise ValueError(velocity_cluster + " not found in column name: " + velocity_cluster_key)
            cells = adata.obs.loc[adata.obs[velocity_cluster_key] == velocity_cluster].index
            cells_idx = np.where(adata.obs_names.isin(cells))[0].tolist()
        
            X_grid, V_grid = get_embedding_subset_grid(adata = adata,
                                                       cells_idx = cells_idx,
                                                       basis = basis, vkey = vkey,
                                                       adjust_for_stream = adjust_for_stream)
            
            entrain_dat = adata.uns["entrain_velocity_ligands"]
            ligand_importances = entrain_dat["vclust_ligand_importances"][velocity_cluster].fillna(0)
            ligand_importances.index = entrain_dat["vclust_ligand_importances"]["ligand"]
            ligand_importances = ligand_importances.sort_values(ascending=False)
            ligands = ligand_importances.index.values
            
            influence_col = "influence" + velocity_cluster
            adata.obs[influence_col] = get_sender_influence(adata, ligands)

            if vector_type == "grid":
                ax.append(
                    scv.pl.velocity_embedding_grid(adata, 
                                                   color = influence_col,
                                                   dpi=dpi, 
                                                   size=size,
                                                   arrow_length=arrow_length,
                                                   arrow_size=arrow_size,
                                                   color_map=sender_palette,
                                                   density=density,
                                                   arrow_color=arrow_color, 
                                                   ax=plt.subplot(gs[gs_row,gs_col]),
                                                   X_grid = X_grid,
                                                   V_grid = V_grid,
                                                   title = velocity_cluster,
                                                   autoscale=True,     
                                                   frameon=frameon,
                                                   show=False)
                    )
            elif vector_type == "stream":
                ax.append(
                    scv.pl.velocity_embedding_stream(adata, 
                                                   color = influence_col,
                                                   dpi=dpi, 
                                                   size=size,
                                                   arrow_size=arrow_size,
                                                   color_map=sender_palette,
                                                   density=density,
                                                   arrow_color=arrow_color, 
                                                   ax=plt.subplot(gs[gs_row,gs_col]),
                                                   X_grid = X_grid,
                                                   V_grid = V_grid,
                                                   title = velocity_cluster,
                                                   frameon=frameon,
                                                   show=False)
                    ) 

            ax[0].figure.savefig(filename, dpi=dpi)

    print("Saved at "+ os.getcwd() + "/" + filename)

    return(None)


def plot_velocity_ligands_python(adata,
                          color = None,
                          top_n_ligands=5,
                          vector_type = "stream",
                          velocity_cluster_key="velocity_clusters",
                          velocity_clusters = None,
                          vkey="velocity",
                          basis="umap",
                          velocity_cluster_palette = "Spectral",
                          dpi=300,
                          arrow_length=1,
                          linewidth_scale = 2,
                          density=3,
                          size = 100,
                          alpha=0.1,
                          arrow_size=1,
                          figsize=(16,9),
                          label_fontsize = 15,
                          cell_palette=None,
                          signif_cluster_pct_cutoff = 0.01,
                          plot_output_path="entrain_plot.png",
                          plot_negative_VE = False,
                          **kwargs):

    import matplotlib.pyplot as plt
    import matplotlib as mpl
    from matplotlib import cm
    from adjustText import adjust_text
    if type(adata) == str:
        adata = ad.read_h5ad(adata)

    scv.set_figure_params()
    
    top_n_ligands = int(top_n_ligands)  

    if velocity_clusters is not None:
        if np.all(np.isin(velocity_clusters, adata.obs[velocity_cluster_key].unique())):
            velocity_clusters = velocity_clusters
        else:
            raise ValueError(velocity_clusters + "not found in values of column: " + velocity_cluster_key)
    elif signif_cluster_pct_cutoff > 0.0:
        vcluster_counts = adata.obs[velocity_cluster_key].value_counts()
        velocity_clusters = vcluster_counts[vcluster_counts > (signif_cluster_pct_cutoff * adata.n_obs)].index.to_numpy()
    else:    
        velocity_clusters = adata.obs[velocity_cluster_key].unique()          

    if plot_negative_VE == False:
        var_exp_df =  adata.uns["entrain_velocity_ligands"]["variance_explained"]
        var_exp_df.columns = [velocity_cluster_key, "var_exp"]
        positive_VE_idx = var_exp_df["var_exp"].astype("float") > 0
        positive_ve_clusters = var_exp_df.loc[positive_VE_idx, velocity_cluster_key].tolist()
        velocity_clusters = np.intersect1d(velocity_clusters, positive_ve_clusters)
        if len(velocity_clusters) == 0 :
            raise ValueError("None of the velocity clusters had a positive VE. If you wish to visualize negative VE clusters, set plot_negative_VE = True.")

    label_df = get_velocity_cluster_label_positions(adata = adata, 
                                                    velocity_clusters = velocity_clusters,
                                                    velocity_cluster_key = velocity_cluster_key,
                                                    top_n_ligands = top_n_ligands,
                                                    figsize = figsize,
                                                    dpi=dpi
                                                    )
    plt.clf()

    num_velo_clusters = len(velocity_clusters)  

    emb_x_range = adata.obsm[f"X_{basis}"][:,0].max() - adata.obsm[f"X_{basis}"][:,0].min()
    emb_y_range = adata.obsm[f"X_{basis}"][:,1].max() - adata.obsm[f"X_{basis}"][:,1].min()      

    ax = plt.axes()

    # arrow colors
    if type(velocity_cluster_palette) == str:
        if velocity_cluster_palette in mpl.colormaps():
            arrow_colormap = cm.get_cmap(velocity_cluster_palette, num_velo_clusters)
            arrow_colors = [arrow_colormap(float(i) / num_velo_clusters) for i in range(num_velo_clusters)]
        elif velocity_cluster_palette in mpl.colors.CSS4_COLORS.keys():
            arrow_colors = [velocity_cluster_palette for i in range(num_velo_clusters)]
        else:
            raise ValueError("velocity_cluster_palette argument does not match a matplotlib color name. Valid names are: " + str(mpl.colors.CSS4_COLORS.keys()))
    elif type(velocity_cluster_palette) == list:
        arrow_colors = velocity_cluster_palette
        if len(velocity_cluster_palette) != num_velo_clusters:
            raise ValueError("Number of arrow colors: " + str(arrow_colors) + " does not match the number of velocity clusters being plotted: " + str(velocity_clusters))
    
    adjust_for_stream = True if vector_type=="stream" else False
    for i,velocity_cluster in enumerate(velocity_clusters):
        cells = adata.obs.loc[adata.obs[velocity_cluster_key] == velocity_cluster].index
        cells_idx = np.where(adata.obs_names.isin(cells))[0].tolist()


        if i>0:
            size = 0 # Plot scatter points (cells, not vectors) once.

        if vector_type=="stream": 
                    
            X_grid, V_grid = get_embedding_subset_grid(adata = adata,
                                                       cells_idx = cells_idx,
                                                       basis = basis, vkey = vkey,
                                                       adjust_for_stream = adjust_for_stream)
            
            lengths = np.sqrt((V_grid ** 2).sum(0))
            linewidth = 1 if linewidth_scale is None else linewidth_scale
            linewidth *= 2 * lengths / lengths[~np.isnan(lengths)].max()
            
            grid_x_range = X_grid[0,:].max() - X_grid[0,:].min()
            grid_y_range = X_grid[1,:].max() - X_grid[1,:].min()

            # maintains consistent arrow plot density across clusters.
            density_x = density * (grid_x_range/emb_x_range) 
            density_y = density * (grid_x_range/emb_y_range)
            plot = scv.pl.velocity_embedding_stream(adata, 
                                           color = color,
                                           alpha = alpha,
                                           X_grid = X_grid,
                                           V_grid = V_grid,
                                           dpi = dpi, 
                                           size = size,
                                           linewidth = linewidth,
                                           density=(density_x, density_y),
                                           arrow_size = arrow_size,
                                           palette = cell_palette,
                                           arrow_color = arrow_colors[i], 
                                           ax = ax,
                                           figsize = figsize,
                                           legend_loc = 'none',
                                           show = False, **kwargs)
        elif vector_type == "grid":
            
            
            X_grid, V_grid = get_embedding_subset_grid(adata = adata,
                                                       density = density,
                                                       cells_idx = cells_idx,
                                                       basis = basis, vkey = vkey,
                                                       adjust_for_stream = adjust_for_stream)
            
            grid_x_range = X_grid[0,:].max() - X_grid[0,:].min()
            grid_y_range = X_grid[1,:].max() - X_grid[1,:].min()
    
            # maintains consistent arrow plot density across clusters.
            density_x = density * (grid_x_range/emb_x_range) 

            plot = scv.pl.velocity_embedding_grid(adata, 
                                           color = color,
                                           alpha=alpha,
                                           X_grid = X_grid,
                                           V_grid = V_grid,
                                           dpi = dpi, 
                                           size = size,
                                           density=density_x,
                                           arrow_size = arrow_size,
                                           palette = cell_palette,
                                           arrow_color = arrow_colors[i], 
                                           ax = ax,
                                           figsize = figsize,
                                           legend_loc = 'none',
                                           show = False, **kwargs)
        else:
            raise ValueError("vector_type needs to be one of \"stream\" or \"grid\". ")

    for t in plot.texts:
        t.set_visible(False)

    if top_n_ligands > 0:
        text = []
        for i in range(label_df.shape[0]):
            t = plot.text(x=label_df.loc[i, "pos_x"],
                          y=label_df.loc[i, "pos_y"],
                          s=label_df.loc[i, "label"],
                          bbox=dict(facecolor='white', alpha=0.7),
                          fontsize=label_fontsize)

            text.append(t)
        adjust_text(text)
    fig = plot.get_figure()
    fig.savefig(plot_output_path)
    return(None)

def get_velocity_cluster_label_positions(adata,
                                         velocity_clusters,
                                         velocity_cluster_key,
                                         top_n_ligands,
                                         figsize,
                                         dpi):
    
    import matplotlib as mpl
    labels_plot = scv.pl.velocity_embedding_stream(adata, 
                                       color = velocity_cluster_key,
                                       dpi=dpi, 
                                       figsize=figsize,
                                       legend_loc = "on data",
                                       show=False)
    
    if  adata.uns["entrain_velocity_ligands"] is not None:
        try: 
            entrain_dat = adata.uns["entrain_velocity_ligands"]
            vclust_ligand_importances = entrain_dat["vclust_ligand_importances"]
            vclust_likelihoods = entrain_dat["vclust_gene_likelihoods"]
            vclust_var_exp = entrain_dat["variance_explained"]
            vclust_var_exp.columns = [velocity_cluster_key, "var_exp"]
        except: 
            raise ValueError("adata.uns[\"entrain_velocity_ligands\" has unrecognized names in it's structure. Did Entrain run correctly or have you changed list element names?")
    else:
        raise ValueError("adata.uns[\"vclust_ligand_importances\"] not found. Have you run entrain_velocity on this dataset yet?")

    if velocity_clusters is not None:
        velocity_clusters = velocity_clusters 
    else:
        velocity_clusters = adata.obs[velocity_cluster_key].unique()

    text_labels = [c for c in labels_plot.get_children() if isinstance(c, mpl.text.Text)]
    cluster_label_positions = [c.get_position()
                               for c in text_labels if c.get_text() in velocity_clusters]
    clusters_labels = [c.get_text()
                       for c in text_labels if c.get_text() in velocity_clusters]
    label_df = pd.DataFrame(
        clusters_labels, cluster_label_positions).reset_index()
    label_df[["pos_x", "pos_y"]] = label_df.iloc[:, 0].apply(
        lambda x: pd.Series([x[0], x[1]]))
    label_df["pos_y"] = label_df["pos_y"] - 1
    label_df = label_df.drop("index", 1)
    label_df = label_df.rename({0: velocity_cluster_key}, axis=1)
    label_df = label_df.merge(vclust_var_exp, on = velocity_cluster_key)
    
    temp = []
    for clust in clusters_labels:
        ligands = vclust_ligand_importances.loc[:, ["ligand", clust]].sort_values(by=clust, ascending=False)
        top_ligands = ligands.head(top_n_ligands)["ligand"].values.astype(str).tolist()
        top_ligands.append(clust)
        temp.append(top_ligands)
    top_ligand_df = pd.DataFrame(temp)
    vclust_idx = len(top_ligand_df.columns)-1
    top_ligand_df.rename(columns= {top_ligand_df.columns[vclust_idx] : velocity_cluster_key}, inplace=True)
    label_df = label_df.merge(top_ligand_df, on=velocity_cluster_key)
    label_df["var_exp"] = label_df["var_exp"].astype("float")

    label_df["var_exp"] = round(
        label_df.loc[:, "var_exp"].multiply(100.0), 1).astype(str) + "% V.E."

    label_df["label"] = label_df.iloc[:, 4:].apply(
        lambda x: ', \n'.join(x), axis=1) + " \n" + label_df["var_exp"]
    
    
    return(label_df)

def plot_velocity_clusters_python(adata_clustered,
                                  plot_file,
                                  velocity_cluster_key,
                                  vector_type="stream", palette = None, **kwargs):

    if type(adata_clustered) == str:
        adata_clustered = ad.read_h5ad(adata_clustered)
    
    if vector_type == "stream": 
                    
        plot = scv.pl.velocity_embedding_stream(adata_clustered, 
                                           color = velocity_cluster_key,
                                           save = plot_file,
                                           palette = palette,
                                            **kwargs)
    elif vector_type == "grid":
        plot = scv.pl.velocity_embedding_grid(adata_clustered, 
                                           color = velocity_cluster_key,
                                           save = plot_file,
                                           palette = palette,
                                            **kwargs)
    else:
        raise ValueError("vector_type needs to be one of \"stream\" or \"grid\". ")

    return None


def get_sender_influence(adata, ligands, top_n_ligands = 3):
    sender_ligand_X = adata[:,ligands[0:top_n_ligands]].to_df()
    sender_influence = np.array(sender_ligand_X.sum(axis=1))
    sender_influence_col= pd.Series(sender_influence)
    sender_influence_col.index = adata.obs_names
    
    return(sender_influence_col)
    

def get_embedding_subset_grid(adata, cells_idx, vkey, basis, adjust_for_stream):
    from scvelo.plotting.velocity_embedding_grid import compute_velocity_on_grid
    obsm = adata.obsm
    
    if len(cells_idx) < 51:
        raise ValueError("Too few (less than 50 cells) found in velocity cluster. Ensure your clusters have more than 50 cells or exclude them from analysis with velocity_clusters argument.")
    X_emb = np.array(obsm[f"X_{basis}"][cells_idx,:])
    V_emb = np.array(obsm[f"{vkey}_{basis}"][cells_idx, :])
    
    X_grid, V_grid = compute_velocity_on_grid(
                    X_emb=X_emb,
                    V_emb=V_emb,
                    adjust_for_stream=adjust_for_stream
                    )
    return(X_grid, V_grid)

def get_adata_obs_names(adata):

    if type(adata) == str:
        adata = ad.read_h5ad(adata)

    return(adata.obs_names.values)     




import anndata as ad
import scvelo as scv
import scanpy as sc
import squidpy as sq
import numpy as np
import pandas as pd
import tangram as tg
import matplotlib.pyplot as plt
from pypath.utils import homology
from omnipath.interactions import import_intercell_network
from pypath.utils import mapping
import math, sys, os
import rpy2.robjects as robjects
from rpy2.robjects import numpy2ri
from rpy2.robjects import pandas2ri
from rpy2.robjects import default_converter
from .entrain_scvelo import cluster_velocities
from .entrain_scvelo import recover_dynamics_clusters
from sklearn.preprocessing import MinMaxScaler


def get_velocity_ligands_spatial(
        adata,
        adata_st,
        velocity_cluster_key = "velocity_clusters",
        n_top_genes = 500,
        adata_env=None,
        vclusters = None,
        ligand_target_matrix = None,
        lr_network = None,
        tangram_result_column = "velocity_label_transfer",
        expression_proportion_cutoff=0.05,
        organism="human",
        n_jobs = 4,

):

    adata=ad.read_h5ad(adata) if type(adata) == str else adata
    adata_st=ad.read_h5ad(adata_st) if type(adata_st) == str else adata_st

        # step1 cluster velocities.
    
    if velocity_cluster_key not in adata.obs.columns.values:
        adata = cluster_velocities(adata, cluster_key = velocity_cluster_key)
    
        #step 2 get dynamics
    if "velocity_cluster_likelihoods" not in adata.uns.keys():
        adata = recover_dynamics_clusters(adata, 
                                     return_adata=True,
                                     cluster_key=velocity_cluster_key,
                                     n_jobs = n_jobs,
                                     n_top_genes = n_top_genes)

        # step 3 transfer velocity labels to spatial.
    if tangram_result_column not in adata_st.obs.columns.values:
        adata_st = velocity_label_transfer(adata,
                                    adata_st,
                                    velocity_cluster_key = velocity_cluster_key,
                                    tangram_result_column = tangram_result_column)
    
    if vclusters is None:
        vclusters = adata_st.obs[tangram_result_column].unique()
        
        # get lr pairs expressed with spatial data.
    spatially_interacting_ligands = get_spatially_interacting_ligands(
                                                    adata_st,
                                                    organism=organism,
                                                    vclusters_of_interest=vclusters,
                                                    cluster_key=tangram_result_column,
                                                    expression_proportion_cutoff = expression_proportion_cutoff)
    
    adata.uns["entrain"] = {}
    adata.uns["entrain"]["spatially_interacting_ligands"] = spatially_interacting_ligands
    
    if ligand_target_matrix is None:
        raise(ValueError("Please supply a ligand target matrix"))
    elif type(ligand_target_matrix) == str:
            ligand_target_matrix = pd.read_csv(ligand_target_matrix, index_col=0)
    result_ligands = velocity_ligands(adata,
                                              vclusters_of_interest = vclusters,
                                              ligand_target_matrix = ligand_target_matrix,
                                              ligands_key = "spatially_interacting_ligands")
    adata.uns["entrain_velocity_ligands"] = result_ligands
    return(adata)

def get_velocity_ligands(
        adata,
        velocity_cluster_key = "velocity_clusters",
        n_top_genes = 500,
        vclusters = None,
        sender_clusters = None,
        sender_cluster_key = None,
        ligand_target_matrix = None,
        expression_proportion_cutoff=0.05,
        organism="human",
        n_jobs = 4,
):

    adata=ad.read_h5ad(adata) if type(adata) == str else adata

        # step1 cluster velocities.
    
    if velocity_cluster_key not in adata.obs.columns.values:
        adata = cluster_velocities(adata, cluster_key = velocity_cluster_key)
    
        #step 2 get dynamics
    if "velocity_cluster_likelihoods" not in adata.uns.keys():
        adata = recover_dynamics_clusters(adata, 
                                     return_adata=True,
                                     cluster_key=velocity_cluster_key,
                                     n_jobs = n_jobs,
                                     n_top_genes = n_top_genes)

    if vclusters is None:
        vclusters = adata.obs[velocity_cluster_key].unique()
    
    # get lr pairs expressed without spatial data.

    interacting_ligands = get_interacting_ligands(adata,
                                                    organism=organism,
                                                    vclusters_of_interest=vclusters,
                                                    velocity_cluster_key=velocity_cluster_key,
                                                    sender_clusters = sender_clusters,
                                                    sender_cluster_key = sender_cluster_key,
                                                    expression_proportion_cutoff = expression_proportion_cutoff)
    
    adata.uns["entrain"] = {}
    adata.uns["entrain"]["interacting_ligands"] = interacting_ligands
    
    if ligand_target_matrix is None:
        raise(ValueError("Please supply a ligand target matrix"))
    elif type(ligand_target_matrix) == str:
            ligand_target_matrix = pd.read_csv(ligand_target_matrix, index_col=0)

    result_ligands = velocity_ligands(adata,
                                              vclusters_of_interest = vclusters,
                                              ligand_target_matrix = ligand_target_matrix,
                                              ligands_key = "interacting_ligands")
    adata.uns["entrain_velocity_ligands"] = result_ligands
    return(adata)

def get_adjacent_cells(adata, cell_names_oi):
    connectivities = adata.obsp["spatial_connectivities"]

    neighbouring_cells_idx = np.isin(adata.obs_names, cell_names_oi)
    adjacent_edges = connectivities[neighbouring_cells_idx, :]
    cells_interacting_with_clusters_oi_bool = np.array(
        (adjacent_edges).sum(axis=0) > 0)[0]
    return(cells_interacting_with_clusters_oi_bool)


def get_possible_ligrec(adata, organism, intercell_network, ligand_or_receptor):
    
    if ligand_or_receptor == "ligand":
        intercell_column = "source"
        intercell_symbol_column = "genesymbol_intercell_source"
    elif ligand_or_receptor == "receptor":
        intercell_column = "target"
        intercell_symbol_column = "genesymbol_intercell_target"
    else:
        raise(ValueError("ligand_or_receptor must be one of 'ligand' or 'receptor'"))
            
    
    all_possible_ligrec = intercell_network.loc[:, [
        intercell_column, intercell_symbol_column]].drop_duplicates()

    if organism == "mouse":
        ncbi_tax_id = 10090

        possible_ids = homology.translate(
            all_possible_ligrec[intercell_column], ncbi_tax_id)
        
        all_possible_ligrec_array = np.array(list(
            mapping.map_names(possible_ids,
                              id_type="uniprot",
                              target_id_type="genesymbol",
                              ncbi_tax_id=ncbi_tax_id)
        ))
        common_idx = np.isin(all_possible_ligrec_array, adata.var_names)

        all_possible_ligrec_common = all_possible_ligrec_array[common_idx]
        
    elif organism == "human":
        common_idx = np.isin(all_possible_ligrec, adata.var_names)
        all_possible_ligrec_common = all_possible_ligrec[common_idx].iloc[:, 1].values
        
    else:
        raise(ValueError("organism must be one of 'human' or 'mouse'"))
    
    return(all_possible_ligrec_common)


def subset_adata_to_expressed_genes(adata, expression_proportion_cutoff):
    min_cells = int(np.floor(adata.n_obs * expression_proportion_cutoff))
    sc.pp.filter_genes(adata, min_cells=min_cells)
    return adata

def velocity_label_transfer(adata,
                            adata_st,
                            velocity_cluster_key = "velocity_clusters",
                            tangram_result_column = "velocity_label_transfer",
                            plot = None
                            ):
    
        # tg annoyingly sets names to lowercase. Saving here to restore later. 
    adata_st.var["original_name"] = adata_st.var_names.values
    adata.var["original_name"] = adata.var_names.values

    sc.tl.rank_genes_groups(adata,
                            groupby=velocity_cluster_key,
                            use_raw=False)
    
    markers_df = pd.DataFrame(adata.uns["rank_genes_groups"]["names"]).iloc[0:100, :]
    markers = list(np.unique(markers_df.melt().value.values))
    
    tg.pp_adatas(adata, adata_st, genes=markers)

    ad_map = tg.map_cells_to_space(adata, adata_st,
        mode="clusters", cluster_label = velocity_cluster_key,
        density_prior='rna_count_based',
        num_epochs=500,
        # device="cuda:0",
        device='cpu')
    
    tg.project_cell_annotations(ad_map, adata_st, annotation = velocity_cluster_key)
    
    tg_result = adata_st.obsm["tangram_ct_pred"]
    adata_st.obs[tangram_result_column] = tg_result.idxmax(axis=1)
    
    if type(plot) == str: 
        annotation_list = list(adata_st.obsm["tangram_ct_pred"].columns.unique())
        tg.plot_cell_annotation_sc(adata_st, annotation_list,perc=0.02)
        
            # Restore gene names from lowercase.
    adata_st.var.set_index("original_name", drop=True, inplace=True) 
    adata.var.set_index("original_name", drop=True, inplace=True) 

    return(adata_st)


def get_interacting_ligands(adata,
                         organism,
                         vclusters_of_interest=None,
                         velocity_cluster_key=None,
                         sender_cluster_key = None,
                         sender_clusters = None,
                         expression_proportion_cutoff=0.05
                         ):

    sc.set_figure_params(dpi=300, dpi_save=300,
                         format='.svg', figsize=(16, 16))

    transmitter_params = {"categories": "ligand"}
    receiver_params = {"categories": "receptor"}
    
    intercell_network = import_intercell_network(
        transmitter_params=transmitter_params,
        receiver_params=receiver_params)
    
    if type(adata) == str:
        adata = ad.read_h5ad(adata)

    if vclusters_of_interest is None:
        vclusters_of_interest = adata_st.obs[cluster_key].unique().str()
    
    ligands_dict = {}
    for vcluster in vclusters_of_interest:
        ligands = interacting_ligands(adata,
                             velocity_cluster_key = velocity_cluster_key,
                             vclusters_of_interest = vcluster,
                             intercell_network = intercell_network,
                             expression_proportion_cutoff = expression_proportion_cutoff,
                             organism = organism)
        ligands_dict[vcluster] = ligands
    return(ligands_dict)

def interacting_ligands(adata,
                         organism,
                         vclusters_of_interest=None,
                         velocity_cluster_key=None,
                         sender_cluster_key = None,
                         sender_clusters = None,
                         expression_proportion_cutoff=None,
                         intercell_network = None,
                         ):

    mouse_tax_id = 9606
    human_tax_id = 10090

    duplicated_vars = adata.var_names.duplicated().sum()
    duplicated_obs = adata.obs_names.duplicated().sum()
    if (duplicated_vars > 0):
        print("Found " + str(duplicated_vars) + "duplicated gene names.")
        adata.var_names_make_unique()
    if duplicated_obs > 0:
        print("Found " + str(duplicated_obs) + "duplicated cell barcodes.")
        adata.obs_names_make_unique()
    # get active ligands that are expressed in the clusters of interest.
    all_possible_ligands_common = get_possible_ligrec(adata=adata,
                                                      organism=organism,
                                                      intercell_network=intercell_network,
                                                      ligand_or_receptor="ligand")

    if len(all_possible_ligands_common) < 5:
        RuntimeWarning("There are less than 5 common ligand genes between the AnnData object and the intercell network. Are your gene names formatted correctly and have you chosen the right organism?")

    if sender_clusters:
        if sender_cluster_key:
            sender_cell_names_oi = adata.obs[np.isin(
                    adata.obs[sender_cluster_key], sender_clusters)].index.values
        else:
            raise(ValueError("If you're suppling sender_clusters, please supply a sender_cluster_key as well."))
    else:
        sender_cell_names_oi = adata.obs_names.values

    adata_senders = adata[sender_cell_names_oi, all_possible_ligands_common]

    adata_senders = subset_adata_to_expressed_genes(
        adata_senders, expression_proportion_cutoff)
    expressed_ligands = adata_senders.var_names

    # get only those active receptors that are expressed in the velocity clusters.
    all_possible_receptors_common = get_possible_ligrec(adata=adata,
                                                      organism=organism,
                                                      intercell_network=intercell_network,
                                                      ligand_or_receptor="receptor")
    if len(all_possible_receptors_common) < 5:
        RuntimeWarning("There are less than 5 common receptor genes between the AnnData object and the intercell network. Are your gene names formatted correctly and have you chosen the right organism?")

    receiver_cell_names_oi = adata.obs[np.isin(
            adata.obs[velocity_cluster_key], vclusters_of_interest)].index.values
    adata_receptors_present_in_traj = adata[receiver_cell_names_oi,
                                            all_possible_receptors_common]
    adata_receptors_present_in_traj = subset_adata_to_expressed_genes(
        adata_receptors_present_in_traj, expression_proportion_cutoff)
    expressed_receptors = adata_receptors_present_in_traj.var_names
    
    if organism == "mouse":
        expressed_receptors = homology.translate(
            expressed_receptors, mouse_tax_id, id_type="genesymbol")
        expressed_ligands = homology.translate(
            expressed_ligands, mouse_tax_id, id_type="genesymbol")

        # subset interaction network to ligands that have their
        # corresponding receptor expressed in the trajectory.

    if len(expressed_receptors) < 5 or len(expressed_ligands) < 5:
        RuntimeWarning("There are less than 5 expressed ligand or receptor genes found. Are your gene names formatted correctly and have you chosen the right organism?")

    subset_receptors_idx = np.isin(
        intercell_network["genesymbol_intercell_target"].values, list(expressed_receptors))
    intercell_network_subset = intercell_network.iloc[subset_receptors_idx, :]
    subset_ligands_idx = np.isin(
        intercell_network["genesymbol_intercell_source"].values, list(expressed_ligands))
    intercell_network_subset = intercell_network.iloc[subset_ligands_idx, :]

    if organism == "mouse":
        interacting_ligands = list(
            homology.translate(
                intercell_network_subset["genesymbol_intercell_source"], human_tax_id, id_type="genesymbol")
        )
    elif organism == "human":
        interacting_ligands = list(
            intercell_network_subset["genesymbol_intercell_source"]
        )

    return(interacting_ligands)

def get_adjacent_ligands(adata,
                         organism,
                         intercell_network,
                         cell_names_oi=None,
                         clusters_oi=None,
                         cluster_key=None,
                         expression_proportion_cutoff=0.05
                         ):

    mouse_tax_id = 9606
    human_tax_id = 10090

    sc.set_figure_params(dpi=300, dpi_save=300,
                         format='.svg', figsize=(16, 16))


    if type(adata) == str:
        adata = ad.read_h5ad(adata)

    sq.gr.spatial_neighbors(adata)

    duplicated_vars = adata.var_names.duplicated().sum()
    duplicated_obs = adata.obs_names.duplicated().sum()
    if (duplicated_vars > 0):
        print("Found " + str(duplicated_vars) + "duplicated gene names.")
        adata.var_names_make_unique()
    if duplicated_obs > 0:
        print("Found " + str(duplicated_obs) + "duplicated cell barcodes.")
        adata.obs_names_make_unique()
    # get active ligands that are expressed proximal to the trajectory.
    if cell_names_oi is None:
        cell_names_oi = adata.obs[np.isin(
            adata.obs[cluster_key], clusters_oi)].index.values

    all_possible_ligands_common = get_possible_ligrec(adata=adata,
                                                      organism=organism,
                                                      intercell_network=intercell_network,
                                                      ligand_or_receptor="ligand")
    
    cells_interacting_with_traj_idx = get_adjacent_cells(
        adata, cell_names_oi=cell_names_oi)
    adata_ligands_interacting_with_traj = adata[cells_interacting_with_traj_idx,
                                                all_possible_ligands_common]
    adata_ligands_interacting_with_traj = subset_adata_to_expressed_genes(
        adata_ligands_interacting_with_traj, expression_proportion_cutoff)
    expressed_ligands = adata_ligands_interacting_with_traj.var_names

    # get active receptors rhat are expressed within the trajectory
    all_possible_receptors_common = get_possible_ligrec(adata=adata,
                                                      organism=organism,
                                                      intercell_network=intercell_network,
                                                      ligand_or_receptor="receptor")
    if len(all_possible_receptors_common) < 5:
        RuntimeWarning("There are less than 5 common receptor genes between the AnnData object and the intercell network. Are your gene names formatted correctly and have you chosen the right organism?")

    adata_receptors_present_in_traj = adata[cell_names_oi,
                                            all_possible_receptors_common]
    adata_receptors_present_in_traj = subset_adata_to_expressed_genes(
        adata_receptors_present_in_traj, expression_proportion_cutoff)
    expressed_receptors = adata_receptors_present_in_traj.var_names
    
    if organism == "mouse":
        expressed_receptors = homology.translate(
            expressed_receptors, mouse_tax_id, id_type="genesymbol")
        expressed_ligands = homology.translate(
            expressed_ligands, mouse_tax_id, id_type="genesymbol")

        # subset interaction network to ligands that are both
        # expressed proximal to the trajectory and have their
        # corresponding receptor expressed in the trajectory.
    subset_receptors_idx = np.isin(
        intercell_network["genesymbol_intercell_target"].values, list(expressed_receptors))
    intercell_network_subset = intercell_network.iloc[subset_receptors_idx, :]
    subset_ligands_idx = np.isin(
        intercell_network["genesymbol_intercell_source"].values, list(expressed_ligands))
    intercell_network_subset = intercell_network.iloc[subset_ligands_idx, :]

    if organism == "mouse":
        adjacent_ligands = list(
            homology.translate(
                intercell_network_subset["genesymbol_intercell_source"], human_tax_id, id_type="genesymbol")
        )
    elif organism == "human":
        adjacent_ligands = list(
            intercell_network_subset["genesymbol_intercell_source"]
        )

    return(adjacent_ligands)

def get_spatially_interacting_ligands(adata_st,
                         organism,
                         vclusters_of_interest,
                         cluster_key=None,
                         expression_proportion_cutoff = None):
    
    transmitter_params = {"categories": "ligand"}
    receiver_params = {"categories": "receptor"}
    
    intercell_network = import_intercell_network(
        transmitter_params=transmitter_params,
        receiver_params=receiver_params)
    
    if vclusters_of_interest is None:
        vclusters_of_interest = adata_st.obs[cluster_key].unique().str()
    
    ligands_dict = {}
    for vcluster in vclusters_of_interest:
        adjacent_ligands = get_adjacent_ligands(adata_st,
                             cluster_key = cluster_key,
                             clusters_oi = vcluster,
                             intercell_network = intercell_network,
                             expression_proportion_cutoff = expression_proportion_cutoff,
                             cell_names_oi = None,
                             organism = organism)
        ligands_dict[vcluster] = adjacent_ligands
    return(ligands_dict)

def velocity_ligands(adata,
                             vclusters_of_interest,
                             ligand_target_matrix,
                             ligands_key = "spatially_interacting_ligands"):
    
    ##### R Function importing
    np_cv_rules = default_converter + pandas2ri.converter
    
    rf_regression_str = """
    library("randomForest")
    get_ligand_trajectory_scores_regression <- function(pseudotime_genes, active_ligand_potentials){

    intersection_genes <- base::intersect(rownames(active_ligand_potentials), rownames(pseudotime_genes))
    potentials_intersection <- active_ligand_potentials[intersection_genes,]

    pseudotime_genes_intersection <- pseudotime_genes[intersection_genes,]
    rf <- randomForest::randomForest(x=potentials_intersection, y=pseudotime_genes_intersection, importance=TRUE)

    gene_names = rownames(rf$importance)
    list(genes_input=pseudotime_genes_intersection, gene_names = gene_names, model=rf)
    }
    """
    rf_regression = robjects.r(rf_regression_str)
    #####

    interacting_ligands = adata.uns["entrain"][ligands_key]
    likelihoods_df = adata.uns["velocity_cluster_likelihoods"]
    
    var_exp_df = pd.DataFrame(columns=["vcluster", "var_exp"])
    importance_df = pd.DataFrame(columns=["ligand"]) # coumns = vcluster
    for i, vcluster in enumerate(vclusters_of_interest):
        test_ligands = interacting_ligands[vcluster]
        test_ligands_common = ligand_target_matrix.columns.intersection(test_ligands)
        if len(test_ligands_common) < 10:
            RuntimeWarning("There are less than 10 ligands found in the ligand_target_matrix. Have you supplied the mouse ortholog-converted ligand_target_matrix to human data or vice versa?")

        active_ligand_target = ligand_target_matrix.loc[:, test_ligands_common]
        likelihoods = likelihoods_df[vcluster].dropna()
            
        with (robjects.default_converter + pandas2ri.converter).context():
            rf = rf_regression(pd.DataFrame(likelihoods), active_ligand_target)
        
        var_exp = rf["model"]["rsq"][-1]
        importances_data =  rf["model"]["importance"][:,1] # Column 0 is %IncMSE, Col1 is IncNodePurity
        gene_names = rf["gene_names"]
        importances = pd.DataFrame({"ligand" : gene_names, 
                                    vcluster : importances_data})
        
        var_exp_df.loc[i] = [vcluster, var_exp] 
        importance_df = pd.merge(importance_df, importances, how="outer", left_on="ligand", right_on = "ligand")

    result = {"variance_explained" : var_exp_df,
              "vclust_ligand_importances" : importance_df,
              "vclust_gene_likelihoods" : likelihoods_df
              }  
    
    return(result)


def plot_ligand_importance_heatmap(adata,
                n_top_ligands = 5,
                colorscale = "inferno",
                velocity_clusters = None,
                filename = "velocity_importances.png",
                figsize = (10, 8),
                rescale_columns = True,
                return_plot = True):
    import seaborn as sns

    df = adata.uns["entrain_velocity_ligands"]["vclust_ligand_importances"].set_index("ligand")
    
    # Subset to velocity_clusters
    if velocity_clusters:
        df = df[velocity_clusters]
    
    # Subset to top n_top_ligands for each column
    df = df.apply(lambda x: x.nlargest(n_top_ligands))
    
    if rescale_columns:
        from sklearn.preprocessing import MinMaxScaler
        # Rescale columns to be between 0 and 1
        scaler = MinMaxScaler()
        df = pd.DataFrame(scaler.fit_transform(df), columns=df.columns, index=df.index)
    df=df.fillna(0)

    # Plot heatmap
    fig, ax = plt.subplots(figsize=(10, 8))
    sns.heatmap(df, cmap=colorscale, ax=ax)
    
    # Save the plot
    plt.savefig(filename)
    print("Saving under filename: "+ filename)
    if return_plot:
        return plt

def plot_likelihoods_heatmap(adata,
                n_top_genes = 10,
                colorscale = "viridis",
                velocity_clusters = None,
                filename = "velocity_likelihoods.png",
                figsize = (10, 8),
                rescale_columns = True,
                return_plot = True):
    from sklearn.preprocessing import MinMaxScaler
    import seaborn as sns

    df = adata.uns["entrain_velocity_ligands"]["vclust_gene_likelihoods"]
    
    # Subset to velocity_clusters
    if velocity_clusters:
        df = df[velocity_clusters]
    
    # Subset to top n_top_ligands for each column
    df = df.apply(lambda x: x.nlargest(n_top_genes))
    
    if rescale_columns:
        # Rescale columns to be between 0 and 1
        scaler = MinMaxScaler()
        df = pd.DataFrame(scaler.fit_transform(df), columns=df.columns, index=df.index)
    df=df.fillna(0)

    # Plot heatmap
    fig, ax = plt.subplots(figsize=(10, 8))
    sns.heatmap(df, cmap=colorscale, ax=ax)
    
    # Save the plot
    plt.savefig(filename)
    print("Saving under filename: "+ filename)
    if return_plot:
        return plt        

def plot_ligand_targets_heatmap(
    adata,
    ligand_target_matrix,
    velocity_cluster,
    colorscale = "cividis",
    n_top_ligands = 5,
    n_top_genes = 10,
    filename = "ligand_target_heatmap.png",
    figsize = (10, 4),
    return_plot=True,
    rescale = False,
    coord_flip = True):
    from sklearn.preprocessing import MinMaxScaler
    import seaborn as sns
    df = pd.DataFrame(adata.uns["entrain_velocity_ligands"]["vclust_ligand_importances"].set_index("ligand"))
            
    df = df[velocity_cluster]

    # Get top n largest ligands
    top_ligands = df.nlargest(n_top_ligands).index

    #Get top genes for each ligand
    top_genes = ligand_target_matrix.loc[top_ligands].apply(lambda x: x.nlargest(n_top_genes).index, axis=1)
    top_genes = pd.unique(top_genes.explode())

    # Subset the ligand_target_matrix to the top ligands and top genes
    ligand_target_df = ligand_target_matrix.loc[top_genes, top_ligands]

    # Replace NaN values with 0
    ligand_target_df = ligand_target_df.fillna(0)

    # Rescale to be between 0 and 1. Default False because it is usually useful to compare gene regulatory potentials across different ligands.
    if rescale:
        scaler = MinMaxScaler()
        ligand_target_df = pd.DataFrame(scaler.fit_transform(ligand_target_df), columns=ligand_target_df.columns, index=ligand_target_df.index)
        
    if coord_flip: 
        ligand_target_df = ligand_target_df.transpose() # usually more genes than ligands, so landscape orientation suits.
    # Create a new figure
    fig, ax = plt.subplots(figsize=figsize)
    
    # Plot heatmap
    sns.heatmap(ligand_target_df, cmap=colorscale, ax=ax)
    
    # Save the plot
    fig.savefig(filename)

    if return_plot:
        return plt       



    
import anndata as ad
import scvelo as scv
import scanpy as sc
import squidpy as sq
import numpy as np
import pandas as pd
import tangram as tg
import matplotlib.pyplot as plt
import math, sys, os
from .entrain_spatial import *

def plot_ligand_spatial_expression(
        adata,
        adata_st,
        velocity_clusters,
        top_n_ligands = 5,
        tangram_result_column = "velocity_label_transfer",
        figsize = None,
        nrows=2,
        save = "ligand_spatial_influence.png",
        dpi=300):

    entrain_result = adata.uns["entrain_velocity_ligands"]
    ligand_df = entrain_result["vclust_ligand_importances"]
    var_exp_df = entrain_result["variance_explained"]

    if velocity_clusters is None:
        # get default positive v.e. clusters
        velocity_clusters = var_exp_df.loc[var_exp_df.var_exp > 0, "velocity_clusters"].to_list()
    else:
        if np.all(np.isin(velocity_clusters, adata_st.obs[tangram_result_column].unique())):
            clusters = velocity_clusters
        else:
            raise ValueError(velocity_clusters + "not found in values of column: " + tangram_result_column)
        # set up matplotlib plots

    fig = plt.figure()
    ax = []
    n_clusters = len(velocity_clusters)
    nrows = 1 if n_clusters < 5 else 1 + (math.floor(n_clusters / 4) )
    ncols = n_clusters if n_clusters < 4 else 4
    ax = []
        
    figsize = figsize if figsize is not None else (8*ncols, 8*nrows)
    gs = plt.GridSpec(nrows,
                      ncols,
                      plt.figure(figsize=figsize)
                      ) 
    
    for i,vcluster in enumerate(velocity_clusters):
        vcluster_result = ligand_df.loc[:,["ligand",vcluster]].sort_values(by = vcluster, ascending=False)
        lig_names = vcluster_result.head(top_n_ligands)["ligand"].to_list()
        lig_weights = vcluster_result.head(top_n_ligands)[vcluster].to_numpy()
        
        lig_expr = adata_st[:, lig_names].X.toarray()

        # weighted average expression, weighted by entrain importance.
        if not np.isclose(np.sum(lig_weights), 1):
            lig_weights = lig_weights / np.sum(lig_weights)
        weighted_avg = np.einsum('ij,j->i', lig_expr, lig_weights)
                
        # get cells adjacent to vcluster cells.
        cell_names = adata_st.obs[np.isin(adata_st.obs[tangram_result_column], vcluster)].index.values
        if adata_st.obsp["spatial_connectivities"] is not None:
            adj_cell_idx = get_adjacent_cells(adata_st, cell_names_oi=cell_names)
        else:
            raise ValueError("Please run `squidpy.gr.spatial_neighbors(adata_st)` to calculate spatial connectivities")
        
        weighted_avg[~adj_cell_idx] = 0
        col_name = "ligand_average_expression_" + vcluster
        adata_st.obs[col_name] = weighted_avg
        
        gs_row = math.floor(i/4)
        gs_col = i % 4
        ax.append(
            sc.pl.spatial(adata_st,
                      color = col_name,
                      ax=plt.subplot(gs[gs_row,gs_col]),
                      show=False)[0]
            )

    ax[0].figure.savefig(save, dpi=dpi)
    
    print("Saved at "+ os.getcwd() + "/" + save)
    return ax


# The influence of cell types positioned next to a given velocity cluster, where influence
# is the summed expression of the top n ligands weighted by their variable_importance
# to that velocity cluster.
def plot_influence_proportions(
	adata,
	adata_st,
    tangram_celltype_column,
    tangram_result_column  = 'velocity_label_transfer',
	celltype_key = "broadlabel",
	velocity_cluster_key = "velocity_clusters",
    top_n_ligands=5,
    positive_ve_only=True,
	color = None,
    colormap=None,
    save = "cell_influence_proportions.png",
    dpi=300,

  ):
    if celltype_key not in adata.obs.columns:
        raise ValueError("cluster_key not found in adata.obs.columns")
    if velocity_cluster_key not in adata.obs.columns:
        raise ValueError("velocity_cluster_key not found in adata.obs.columns")
    
    if tangram_celltype_column not in adata_st.obs.columns:
        adata_st = velocity_label_transfer(adata,
                                                       adata_st,
                                                       tangram_result_column = tangram_celltype_column,
                                                       velocity_cluster_key=celltype_key)
    
    if tangram_result_column not in adata_st.obs.columns:
        adata_st = velocity_label_transfer(adata,
                                                       adata_st,
                                                       tangram_result_column = tangram_result_column,
                                                       velocity_cluster_key = velocity_cluster_key)
    
    entrain_result = adata.uns["entrain_velocity_ligands"]
    ve_df = entrain_result["variance_explained"]
    imps_df = entrain_result["vclust_ligand_importances"]
    
    if positive_ve_only:
        vclusters = ve_df.loc[ve_df["var_exp"] > 0, velocity_cluster_key]
    else:
        vclusters = np.sort(adata_st.obs[tangram_result_column].unique())
    
    result = pd.DataFrame(columns = [tangram_celltype_column, "weighted_sum", velocity_cluster_key])
    for vc in vclusters:
        vc_spots = adata_st.obs_names[adata_st.obs[tangram_result_column] == vc]
        adata_vc = adata_st[vc_spots,:]
        
        
        # ligands to consider
        top_ligand_df = imps_df.loc[:,["ligand", vc]].sort_values(vc, ascending=False).head(top_n_ligands)
        top_ligands = top_ligand_df["ligand"]
        top_imps = top_ligand_df[vc].values
        # cells in proximity: 
        if "spatial_connectivities" not in adata_st.obsp.keys():
            try:
                sq.gr.spatial_neighbors(adata_st)
            except:
                raise KeyError("spatial_connectivities not found in adata_st.obsp. Please run squidpy.gr.spatial_neighbors() on adata_st")          
                
        adj_cells_bool = get_adjacent_cells(adata_st, vc_spots)
        adj_cells_names = adata_st.obs_names[adj_cells_bool]
        adj_ligand_mtrx = adata_st[adj_cells_names, top_ligands].X.todense()
        summed_avg_expr = np.multiply(adj_ligand_mtrx, top_imps).sum(axis=1)
        
        summed_avg_expr_df = pd.DataFrame(summed_avg_expr)
        summed_avg_expr_df.index = adj_cells_names
        summed_avg_expr_df.columns = [vc]
        
        df = pd.merge(summed_avg_expr_df, adata_st.obs[tangram_celltype_column],
                 how="left",
                 left_index=True, right_index=True)
        
        df_wsum = df.groupby(tangram_celltype_column)[vc].sum().reset_index()
        df_wsum.columns = [tangram_celltype_column, 'weighted_sum']
        df_wsum["weighted_sum"] = df_wsum["weighted_sum"]/np.sum(df_wsum["weighted_sum"])
        df_wsum[velocity_cluster_key] = vc
        
        result = pd.concat([result, df_wsum])
        
    pivot_df = result.pivot_table(values='weighted_sum', index=velocity_cluster_key, columns=tangram_celltype_column, aggfunc='sum')
    pivot_df = pivot_df.fillna(0)
    ax=pivot_df.plot(kind='barh',
                  stacked=True,
                  figsize=(10,7),
                  width = 0.8,
                  color=color,
                  colormap=colormap)
    plt.title('Cell Type Influence on Velocities')
    plt.xlabel('Velocity Clusters')
    plt.ylabel('Cell Type Contributions')
    ax.figure.savefig(save, dpi=dpi)
    print("Saved at "+ os.getcwd() + "/" + save)
    
    return ax

